'' **********************
'' *   Chip-8 Emulator  *
'' **********************
''(c)2009 Andy Schenk  see MIT licence at end

' Chip8 is an early Game Programming Language
' see Wikipedia "CHIP-8" and "SCHIP"

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  'SND = 0'10    '<-- Sound Pin

VAR
  long rndm, time, timrate, speed
  long exit
  long addr[16]   'game pointers
  long speedIndex[16] 'game speed delays array (JW correction)

  long sp,pc      'Chip8 Registers
  long stack[16]
  long IR         'Address Register,
  byte V[16]      '8bit Registers V0..VF
  byte T[2]       'Delay Timer Registers


OBJ
  di : "Chip8_TV"
  kb : "bel-keyb"

PUB Main  | i,p
  di.start(20,1)      '<-- TV-basepin and Mode (0=PAL 1=NTSC)
  kb.start(17,16)        '<--- Keyboard Basepin
  rndm := cnt
  timrate := clkfreq/60
  'dira[SND] := 1

  repeat
    di.colmode(2,1)
    di.cls
    di.at(0,0)
    p := @game1
    repeat i from 1 to 9    'Menu
      if long[p] == 0
        quit
      addr[i] := p+16
      speedIndex[i] := long[p+4]   '(JW correction)
      di.hex(i,1)
      di.str(p+8)
      p += long[p]
      if (i&1) == 0
        di.at(0,i*9)
    di.str(string("q Quit"))
    i := kb.newkey          'Wait for Key
    if i=="q"
      reboot
    i -= "0"
    if addr[i] == 0
      next
    bytemove(@mem+512,addr[i],3076)  'load Game
    speed := speedIndex[i]           'get speed delay (JW correction)
    di.cls                           'show screen
    di.box(0,0,31,34)
    di.box(96,0,30,34)
    di.plot(31,0)
    di.drawto(96,0)
    di.plot(31,34)
    di.drawto(96,34)
    di.at(4,36)
    di.str(string("123C|1 2 3 4"))
    di.at(4,50)
    di.str(string("456D| Q W E R"))
    di.at(4,64)
    di.str(string("789E|  A S D F"))
    di.at(4,78)
    di.str(string("A0BF|   Z X C V"))
    di.colmode(1,2)

    pc := $200                       'run emulation loop
    sp := 0
    time := cnt
    repeat
      do_instr(mem[pc]<<8 + mem[pc+1])
    until kb.keystate($CB)           'until ESC key

PUB do_instr(code) | x,y,d       'CPU emulation
{
  dbg(0,pc,3)
  dbg(32,code,4)
  kb.getkey
'}
  x := code >> 8 & $F
  y := code >> 4 & $F
  case code>>12
    0: case code&$FF
'         $E0: di.cls             'clear screen
         $E0: di.clrbox(32,1,64,32) 'clear screen
         $EE: pc := stack[--sp]  'return
    1: pc := code & $FFF - 2     'jump to NNN
    2: stack[sp++] := pc         'call NNN
       pc := code & $FFF - 2
    3: if V[x] == code&$FF       'skip if Vx==KK
         pc+=2
    4: if V[x] <> code&$FF       'skip if Vx<>KK
         pc+=2
    5: if V[x] == V[y]           'skip if Vx==Vy
         pc+=2
    6: V[x] := code & $FF        'Vx=KK
    7: V[x] += code & $FF        'Vx=Vx+KK
    8: case code & $F
         0: V[x] := V[y]         'Vx=Vy
         1: V[x] |= V[y]         'Vx=Vx OR Vy
         2: V[x] &= V[y]         'Vx=Vx AND Vy
         3: V[x] ^= V[y]         'Vx=Vx XOR Vy
         4: d := V[x] + V[y]     'Vx=Vx+Vy VF=Carry
            V[x] := d & $FF
            V[$F] := d>>8
         5: d := V[x] - V[y]     'Vx=Vx-Vy VF= !Borrow
            V[x] := d & $FF
            V[$F] := d>>8 + 1
         6: V[$F] := V[x]&1      'Vx SHR 1 VF=Carry
            V[x] >>= 1
         7: d := V[y] - V[x]     'Vx=Vy-Vx VF= !Borrow
            V[x] := d & $FF
            V[$F] := d>>8 + 1
         $E: V[$F] := V[x]>>7&1  'Vx SHL 1 VF=Carry
             V[x] <<= 1
    9: if V[x] <> V[y]           'skip if Vx<>Vy
         pc+=2                   '(JW correction)
    $A: IR := code & $FFF        'I=NNN
    $B: pc := (code & $FFF) + V[0] 'jump to NNN+V0
    $C: V[x] := ?rndm & (code&$FF) 'Vx = random & NN
    $D: V[$F] := drawSprite(V[x],V[y],code&$F)  'Sprite 8*N
    $E: case code&$FF
          $9E: if key(V[x])      'skip if key Vx pressed
                 pc+=2
          $A1: ifnot key(V[x])   'skip if key Vx not pressed
                 pc+=2
    $F: case code&$FF
          $07: V[x] := T[0]       'Vx=Delay Timer
          $0A: V[x] := waitkey    'Wait for Key
          $15: T[0] := V[x]       'Delay Timer = Vx
          $18: T[1] := V[x]       'Sound Timer = Vx
          $1E: IR += V[x]         'I=I+Vx
          $29: IR := @font-@mem + V[x]*5  'Font pointer to I
          $33: mem[IR] := V[x]/100 'BCD to mem[I..I+2]
               mem[IR+1] := (V[x]/10)//10
               mem[IR+2] := V[x]//10
          $55: repeat y from 0 to x  'save V0..Vx to Mem[I]
                 mem[IR+y] := V[y]
          $65: repeat y from 0 to x  'load V0..Vx from Mem[I]
                 V[y] := mem[IR+y]

  if T[1] > 0                     'Sound Timer
    '!outa[SND]                    'Beep
    T[1]--
  if cnt - time > timrate         'Timers
    time := cnt
    if T[0] > 0   'Delay Timer
      T[0]--

  if speed>0
    waitcnt(clkfreq/speed + cnt)

  pc+=2


PUB key(n) : p   '' Key n pressed?
  p := kb.keystate(keymap[n&15])

PUB waitkey : k  '' Wait until a Key pressed
  k := kb.newkey
  k := lookdownz( k : "x123qweasdzc4rfv")
'  dbg(0,k,2)

PUB drawSprite(x,y,h) 'draw Sprite at x,y with heigth h
  if h>0
    return di.sprite8(x&$3F+32,y&$1F+1,@mem+IR,h)
  else
    return di.sprite16(x&$7F,y&$3F,@mem+IR)

PRI dbg(x,n,w)
  di.clrbox(x,60,w*8,16)
  di.colmode(1,1)
  di.hex(n,w)
  di.colmode(1,2)


DAT
' keymap:     on Keyboard
'  1 2 3 C  =  1 2 3 4
'  4 5 6 D  =   Q W E R
'  7 8 9 E  =    A S D F
'  A 0 B F  =     Z X C V

keymap  byte "x123qweasdzc4rfv",0,0

mem     byte 0[512]
        byte 0[3584]

font    byte %0100000   '0
        byte %1010000
        byte %1010000
        byte %1010000
        byte %0100000

        byte %0010000   '1
        byte %0110000
        byte %1010000
        byte %0010000
        byte %0010000

        byte %1100000   '2
        byte %0010000
        byte %0100000
        byte %1000000
        byte %1110000

        byte %1100000   '3
        byte %0010000
        byte %1100000
        byte %0010000
        byte %1100000

        byte %1010000   '4
        byte %1010000
        byte %1110000
        byte %0010000
        byte %0010000

        byte %1110000   '5
        byte %1000000
        byte %1100000
        byte %0010000
        byte %1100000

        byte %0100000   '6
        byte %1000000
        byte %1100000
        byte %1010000
        byte %0100000

        byte %1110000   '7
        byte %0010000
        byte %0010000
        byte %0100000
        byte %0100000

        byte %0100000   '8
        byte %1010000
        byte %0100000
        byte %1010000
        byte %0100000

        byte %0100000   '9
        byte %1010000
        byte %0110000
        byte %0010000
        byte %1100000

        byte %0100000   'A
        byte %1010000
        byte %1110000
        byte %1010000
        byte %1010000

        byte %1100000   'B
        byte %1010000
        byte %1100000
        byte %1010000
        byte %1100000

        byte %0110000   'C
        byte %1000000
        byte %1000000
        byte %1000000
        byte %0110000

        byte %1100000   'D
        byte %1010000
        byte %1010000
        byte %1010000
        byte %1100000

        byte %1110000   'E
        byte %1000000
        byte %1100000
        byte %1000000
        byte %1110000

        byte %1110000   'F
        byte %1000000
        byte %1110000
        byte %1000000
        byte %1000000

' Included Games (max. 9)
' --------------
game1   long @game2-@game1, 0    '2 longs: offset, speed
        byte "Blitz  ",0         '8 bytes: name+0
        file "blitz.ch8"         'game binary

game2   long @game3-@game2, 2000
        byte "Blinky ",0
        file "blinky.ch8"

game3   long @game4-@game3, 0
        byte "Guess  ",0
        file "guess.ch8"

game4   long @game5-@game4, 700
        byte "Pong2  ",0
        file "pong2.ch8"

game5   long @game6-@game5, 1500
        byte "Invadr ",0
        file "invaders.ch8"

game6   long @game7-@game6, 2000
        byte "Brix   ",0
        file "brix.ch8"

game7   long @game8-@game7, 1500
        byte "Hidden ",0
        file "hidden.ch8"

game8   long @game9-@game8, 1000
        byte "Tetris ",0
        file "tetris.ch8"

game9   long @gameA-@game9, 1000
        byte "TicTac ",0
        file "tictac.ch8"

gameA   long 0,0


{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
