''****************************************
''*  TV Driver for Chip8 Emulator v0.1   *
''****************************************
'' 128x96 pixel, 4 colors, 3kByte Screenbuffer
'  (c)2009 Andy Schenk, Insonix

CON
  _stack = ($1000 + 100) >> 2   'accomodate display memory and stack

  x_tiles = 8                   'do not change this
  y_tiles = 6

  display_base = $7000          'for TV parameters go to DAT section

  #1, _colmd,_at,_plot,_line,_char,_spr8    'Assembly functions


VAR

  long  cog

  word  screen[x_tiles * y_tiles]
  long  colors[4]


OBJ

  tv    : "tv"

PUB start(basepin, ntsc) : okay | i,k

  'start tv
  tv_screen := @screen
  tv_colors := @colors
  tv_pins := (basepin&$38)<< 1 | (basepin&4==4)&%0101
  tv_mode ^= ntsc
  tv_hx -= ntsc*4
  tv.start(@tv_status)

  'init colors
  repeat i from 0 to 3
    colors[i] := $07050603

  'init tile screen
  repeat i from 0 to tv_hc - 1
    repeat k from 0 to tv_vc - 1
      screen[k * tv_hc + i] := display_base >> 6 + k + i * tv_vc + ((k & $3) << 10)

  'start cog for assembly graphic routines
  okay := cog := cognew(@entry, @cmd) + 1


PUB stop

  tv.stop
  if cog
    cogstop(cog~ -  1)


PUB sprite8(x,y,pntr,h)
' draw a 8*h pixel sprite at x,y
  par1 := x<<16 + y
  par2 := h<<16 + pntr
  do_command(_spr8)
  return (par2<>0)&1   'return 0|1 for collison


PUB sprite16(x,y,pntr)
' not yet implemented

PUB cls
  longfill(display_base, 0, 768)

PUB colmode(col, md)
' set graphic color and mode
  gmode := md
  gcolor := col
  longmove(@par1,@col,2)
  do_command(_colmd)

PUB at(gx, gy)
' set draw pos
  longmove(@par1,@gx,2)
  do_command(_at)


PUB str(stringptr)

'' Print a zero-terminated string

  repeat strsize(stringptr)
    out(byte[stringptr++])


PUB hex(value, digits) | n

'' Print a hexadecimal number

  value <<= (8 - digits) << 2
  repeat digits
    out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))


PUB out(c) | i, k

'' Output a character
''
''     $00 = clear screen
''     $01 = home
''     $08 = backspace
''     $09 = tab (8 spaces per)
''     $0D = return
''  others = printable characters

  case c
    $00: cls
    $01: xcurs := ycurs := 0
    $08: xcurs -= 8
    $09: xcurs := (xcurs / 64)*64 + 64
    $0D: newline
    other:
      par1 := c
      par2 := 256
      do_command(_char)
      xcurs++
      if xcurs > 15
        newline
       

PRI newline
  xcurs := 0
  ycurs += 16
  if ycurs > y_tiles*16
    ycurs -= 16

    
PUB plot(gx, gy)

'' Plot a point at gx,gy

  longmove(@xcurs,@gx,2)
  longmove(@par1,@gx,2)
  do_command(_plot)


PUB drawto(gx, gy)

'' Draw a line from last point to gx,gy

  at(xcurs,ycurs)
  longmove(@par1,@gx,2)
  do_command(_line)
  longmove(@xcurs,@gx,2)

    
PUB box(gx, gy, gw, gh) | i

'' Draw a box at gx,gy with width gw and height gh

  longmove(@xcurs,@gx,2)
  colmode(gcolor,1)
  repeat i from 0 to gh
    at(gx,gy+i)
    line(gx+gw, gy+i)


PUB clrbox(gx, gy, gw, gh) | i,old 

'' Clear the the area from gx,gy with width gw and height gh

  old := gmode
  colmode(gcolor,0)
  
  longmove(@xcurs,@gx,2)
  repeat i from 0 to gh
    at(gx,gy+i)
    line(gx+gw, gy+i)

  colmode(gcolor,old)
  at(gx,gy)
  

PRI line(gx, gy)
' draw line to pos
  longmove(@par1,@gx,2)
  do_command(_line)


PRI do_command(c)
' start assembly command and wait until done
  cmd := c
  repeat while cmd>0

PUB VSync
  repeat until tv_status<>1
  repeat until tv_status==1

  
DAT

'---- TV parameters ---- (PAL)

tv_status               long    0               'status
tv_enable               long    1               'enable
tv_pins                 long    %001_0101       'pins
tv_mode                 long    %0001           'mode
tv_screen               long    0               'screen
tv_colors               long    0               'colors
tv_hc                   long    x_tiles         'hc
tv_vc                   long    y_tiles         'vc
tv_hx                   long    24              'hx
tv_vx                   long    2               'vx
tv_ho                   long    0               'ho
tv_vo                   long    0               'vo
tv_broadcast            long    0               'broadcast
tv_auralcog             long    0               'auralcog


' ---- Assembly Drawing routines ----
                        org     0
entry
                        mov     cmdp,par         'get pointers
                        mov     parx,par
                        mov     pary,par
                        add     parx,#4
                        add     pary,#8
                        
cmdloop                 rdlong  cmd,cmdp
                        tjz     cmd,#cmdloop     'wait for command
                        cmp     cmd,#6      wz
              if_z      jmp     #spriteb
                        cmp     cmd,#1      wz
              if_z      jmp     #colmd
                        cmp     cmd,#2      wz
              if_z      jmp     #atpos
                        cmp     cmd,#3      wz
              if_z      jmp     #setpix
                        cmp     cmd,#4      wz
              if_z      jmp     #lineto
                        cmp     cmd,#5      wz
              if_z      jmp     #print
endcmd                  mov     cmd,#0
                        wrlong  cmd,cmdp
                        jmp     #cmdloop              

colmd                   rdlong  gcolor,parx      'set graphic color and mode
                        rdlong  gmode,pary
                        jmp     #endcmd              

atpos                   rdlong  xcurs,parx       'set drawing location
                        rdlong  ycurs,pary
                        jmp     #endcmd              

setpix                  rdlong  xcurs,parx       'draw pixel
                        rdlong  ycurs,pary
                        call    #setpixel
                        jmp     #endcmd              

lineto                  rdlong  xnew,parx        'draw line to position
                        rdlong  ynew,pary
                        mov     xinc,#1          'bresenham
                        mov     deltx,xnew
                        sub     deltx,xcurs  wc,wz
              if_b      neg     xinc,xinc
              if_b      neg     deltx,deltx
              if_z      mov     xinc,#0     
                        mov     yinc,#1
                        mov     delty,ynew
                        sub     delty,ycurs  wc,wz
              if_b      neg     yinc,yinc
              if_b      neg     delty,delty
              if_z      mov     yinc,#0
                        mov     dist,deltx
                        cmp     deltx,delty  wc
              if_b      mov     dist,delty
                        mov     errx,deltx
                        mov     erry,delty
                        mov     tcnt,dist    wz
              if_z      add     tcnt,#1
'              if_z      jmp     #endcmd
bresloop                call    #setpixel
                        add     errx,deltx
                        add     erry,delty
                        cmp     errx,dist    wc,wz
              if_a      sub     errx,dist
              if_a      add     xcurs,xinc
                        cmp     erry,dist    wc,wz
              if_a      sub     erry,dist
              if_a      add     ycurs,yinc
                        djnz    tcnt,bresloop
                        call    #setpixel
                        jmp     #endcmd              

print                   rdlong  xnew,parx          'draw a character (char,maxX)
                        rdlong  dist,pary          'from ROM-font scaled to 8x16 pixel
                        shr     xnew,#1       wc
                        shl     xnew,#7
                        add     xnew,fontofs
                        mov     delty,#31
:loopy                  rdlong  deltx,xnew
              if_nc     shl     deltx,#1
                        mov     tcnt,#8
:loopx                  test    deltx,#%0010  wz
              if_nz     call    #setpixel
                        shr     deltx,#4
                        add     xcurs,#1
                        djnz    tcnt,#:loopx
                        test    delty,#1      wz
              if_z      add     ycurs,#1
                        sub     xcurs,#8
                        add     xnew,#4
                        djnz    delty,#:loopy     
                        sub     ycurs,#15
                        add     xcurs,#8
                        cmp     xcurs,dist    wc
              if_ae     sub     xcurs,dist
              if_ae     add     ycurs,#16 
                        cmp     ycurs,#192    wc
              if_ae     sub     ycurs,#16
                        jmp     #endcmd              

spriteb                 rdlong  xnew,parx        'draw a 8 bit sprite (X<<16+Y)
                        mov     ycurs,xnew
                        and     ycurs,#$1FF
                        mov     xcurs,xnew
                        shr     xcurs,#16
                        rdlong  dist,pary        '(h<<16+pntr)
                        mov     delty,dist
                        shr     delty,#16        '=h
                        mov     deltx,#0         'collision flag
:looph                  rdbyte  ynew,dist
                        mov     tcnt,#8
                        shl     ynew,#24
:loopx                  shl     ynew,#1  wc
              if_c      call    #setpixel
                        or      deltx,t2
                        add     xcurs,#1
                        djnz    tcnt,#:loopx
                        add     ycurs,#1
                        add     dist,#1
                        sub     xcurs,#8
                        djnz    delty,#:looph
                        wrlong  deltx,pary       'collision to pary
                        jmp     #endcmd

' ---- subroutines ----
setpixel                mov     tx,xcurs         'draw pixel with gmode
                        shr     tx,#4
                        mov     t1,tx
                        shl     tx,#8
                        shl     t1,#7
                        add     tx,t1            '*396
                        mov     ty,ycurs
                        shl     ty,#2
                        add     ty,tx             
                        add     ty,gbase
                        mov     tx,xcurs
                        shl     tx,#1
                        and     tx,#$01E
                        mov     t1,gcolor
                        shl     t1,tx
                        mov     t2,#1 '3
                        shl     t2,tx
                        rdlong  tx,ty
                        cmp     gmode,#0     wz
              if_z      andn    tx,t2
                        cmp     gmode,#1     wz
              if_z      or      tx,t1
                        cmp     gmode,#2     wz
              if_z      and     t2,tx          'old pixel for collision
              if_z      xor     tx,t1
                        wrlong  tx,ty
setpixel_ret            ret

' ----- variables ----
cmd                     long    0
par1                    long    0
par2                    long    0
gcolor                  long    3
gmode                   long    1
gbase                   long    display_base
fontofs                 long    $8000
xcurs                   long    0
ycurs                   long    0

cmdp                    res     1
parx                    res     1
pary                    res     1
t1                      res     1
t2                      res     1
tx                      res     1
ty                      res     1
xnew                    res     1
ynew                    res     1
deltx                   res     1
delty                   res     1
xinc                    res     1
yinc                    res     1
errx                    res     1
erry                    res     1
dist                    res     1
tcnt                    res     1
                        